package com.gigaspaces.ucs.scaleHandler;

import java.net.URI;
import java.net.URISyntaxException;
import java.net.UnknownHostException;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;

import org.openspaces.admin.Admin;
import org.openspaces.admin.AdminFactory;
import org.openspaces.admin.machine.Machine;
import org.openspaces.admin.machine.events.MachineLifecycleEventListener;
import org.openspaces.admin.os.OperatingSystemDetails.NetworkDetails;
import org.openspaces.admin.pu.ProcessingUnit;
import org.openspaces.admin.pu.events.ProcessingUnitRemovedEventListener;
import org.openspaces.grid.esm.ElasticScaleHandler;
import org.openspaces.grid.esm.ElasticScaleHandlerConfig;
import org.openspaces.grid.esm.ElasticScaleHandlerContext;

import com.gigaspaces.ucs.deploy.AgentlessInstaller;
import com.gigaspaces.ucs.deploy.AgentlessInstaller.InstallationDetails;
import com.gigaspaces.ucs.domain.out.LsServer;
import com.gigaspaces.ucs.service.Service;
import com.gigaspaces.ucs.service.UCSServiceException;

/***********
 * Elastic scale handler implementation for the Cisco UCS system.
 * 
 * @author barakme
 *
 */
public class UCSElasticScaleHandler implements ElasticScaleHandler, MachineLifecycleEventListener,
		ProcessingUnitRemovedEventListener {

	private static java.util.logging.Logger logger = java.util.logging.Logger.getLogger(UCSElasticScaleHandler.class
			.getName());

	// UCS Manager Service facade.
	private Service service;
	
	// Manager credentials
	private String username;
	private String password;
	
	// Manager configuration
	private String template;	
	private String organization;
	private String serverNamePrefix;
	private int maxServerIndex = -1;
	private String fullPrefix;
	private String macPoolName;
	private String bladePoolName;
	
	// Gigaspaces properties
	private String group;
	private Admin admin;
	private String locator;
	private Object gridName;
	
	// Agentless installer configuration
	private AgentlessInstaller installer;
	private String sshPassword;
	private String sshUsername;
	private String localDir;
	private String remoteDir;
	private String myIp;

	// Concurrency
	private final Object mutex = new Object();
	private String pendingMachineDN = null;
	private ExecutorService executor;


	/*********
	 * Constructor.
	 */
	public UCSElasticScaleHandler() {
		logger.info("UCSElasticScaleHandler Started");
	}

	@Override
	public boolean accept(final Machine machine) {
		logger.info("UCSElasticScaleHandler: accept: " + machine.getUid());
		return true;
	}

	private void checkCurrentIndex(final List<LsServer> servers) {
		if (maxServerIndex < 0) {

			int valueFound = 0;

			for (final LsServer server : servers) {

				final String dn = server.getDn();
				if (dn.startsWith(fullPrefix)) {
					final String postfix = dn.substring(fullPrefix.length());
					int index = -1;
					try {
						index = Integer.parseInt(postfix);
					} catch (final NumberFormatException e) {
						// ignore, in-case there is a template name collision
						// somewhere
					}
					valueFound = Math.max(valueFound, index);
				}
			}

			this.maxServerIndex = valueFound;
		}
	}

	private LsServer checkForPoweredOffMachine(final List<LsServer> profiles) throws UCSServiceException {
		for (final LsServer profile : profiles) {
			logger.info(profile.getDn() + ": " + profile.getAssocState());

			if (profile.getAssocState().equals("associated")) {
				if (!service.isProfilePoweredOn(profile.getDn())) {
					logger.info("Found profile: " + profile.getDn() + ". Booting it up");
					service.bootServer(profile.getDn());
					return profile;
				} else {
					logger.info(profile.getDn() + ": ON");
				}
			}

		}

		return null;
	}

	private LsServer checkForUnassociatedProfile(final List<LsServer> profiles) throws UCSServiceException {
		for (final LsServer profile : profiles) {
			logger.info(profile.getDn() + ": " + profile.getAssocState());
			// associated
			if (profile.getAssocState().equals("unassociated")) {
				logger.info("Found profile: " + profile.getDn() + ". " + "Associating it with blade from pool "
						+ this.bladePoolName);
				// associate it
				service.associateProfileToBladePool(profile.getDn(), this.bladePoolName);
				// if powered off, boot it
				if (!service.isProfilePoweredOn(profile.getDn())) {
					logger.info("Booting profile: " + profile.getDn() + " up");
					service.bootServer(profile.getDn());
				}
				return profile;
			}

		}

		return null;
	}

	private void cleanUp() {

		// Clean up admin
		logger.info("Shutting down admin API used by handler");
		admin.getMachines().removeLifeycleListener(this);
		admin.getProcessingUnits().getProcessingUnitRemoved().remove(this);
		admin.close();

		// shut down the executor service
		executor.shutdown();

	}

	private InstallationDetails createInstallationDetails() {
		final InstallationDetails details = new InstallationDetails();

		details.setLocalDir(this.localDir);
		details.setRemoteDir(this.remoteDir);
		details.setUsername(this.sshUsername);
		details.setPassword(this.sshPassword);

		details.setLocator(this.locator);
		details.setLus(false);

		details.setAdmin(this.admin);

		return details;
	}

	/*********
	 * Creates a new service profile from the profile template, with a name
	 * according to the configured prefix. If this is the first time that the
	 * template is used in this handler, also scans the given profiles to find
	 * the highest index used in any existing profile name.
	 * 
	 * @param profiles
	 *            the existing profiles.
	 * @return the created server, or null if no template is configured.
	 * @throws UCSServiceException
	 *             In case of an error while communicating with the UCS manager.
	 */
	private LsServer createProfile(final List<LsServer> profiles) throws UCSServiceException {

		if ((this.template != null) && (this.template.length() > 0)) {
			checkCurrentIndex(profiles);

			final int newId = ++maxServerIndex;
			final String serverName = this.serverNamePrefix + newId;

			logger.info("Requesting a new machine from the UCS manager. " + "Template: " + template
					+ ", organization: " + organization + ", name: " + serverName);

			final LsServer newServer = service.startServerFromTemplate(template, organization, this.serverNamePrefix
					+ newId);

			logger.info("New machine boot sequence started. Machine DN is: " + newServer.getDn());
			return newServer;
		}
		return null;

	}

	public Admin getAdmin() {
		return this.admin;
	}	

	public AgentlessInstaller getInstaller() {
		return this.installer;
	}

	private String[] getMACAddressesFromMachine(final Machine machine) {
		final Collection<NetworkDetails> networks = machine.getOperatingSystem().getDetails().getNetworkDetails()
				.values();

		final String[] macAddresses = new String[networks.size()];
		int i = 0;
		for (final NetworkDetails networkDetails : networks) {
			macAddresses[i] = networkDetails.getAddress();
			++i;
		}
		logger
				.info("MAC addresses for machine: " + machine.getHostAddress() + " are: "
						+ Arrays.toString(macAddresses));
		return macAddresses;

	}

	private String getMyIP(final Admin admin) {
		return admin.getElasticServiceManagers().waitForAtLeastOne().getMachine().getHostAddress();
	}

	public String getSshPassword() {
		return this.sshPassword;
	}

	public String getSshUsername() {
		return this.sshUsername;
	}

	/**********
	 * Called by the async scale out task to notify handler that new
	 * scale out tasks can now be performed.
	 * 
	 * @param serverDN the DN of the server that was being started by the async task
	 */
	public void gsaTaskFinished(final String serverDN) {
		logger.info("gsaTaskFinished: " + serverDN);
		synchronized (this.mutex) {
			if ((this.pendingMachineDN != null) && this.pendingMachineDN.equals(serverDN)) {
				logger.info("Pending task for service profile: " + serverDN + " has finished");
				this.pendingMachineDN = null;
			}
		}
	}

	@Override
	public void init(final ElasticScaleHandlerConfig config) {
		logger.info("UCSElasticScaleHandler: init");

		final String uri = config.getProperty("Uri");
		username = config.getProperty("UCSUsername");
		password = config.getProperty("UCSPassword");
		template = config.getProperty("Template");
		organization = config.getProperty("Organization");
		serverNamePrefix = config.getProperty("serverNamePrefix");
		bladePoolName = config.getProperty("BladePoolName");

		sshUsername = config.getProperty("SSHUsername");
		sshPassword = config.getProperty("SSHPassword");

		this.localDir = config.getProperty("localDir");
		this.remoteDir = config.getProperty("remoteDir");

		group = config.getProperty("lookupGroup");
		locator = config.getProperty("locator");
		macPoolName = config.getProperty("macPoolName");

		final String tmpGridName = config.getProperty("gridName");
		if (tmpGridName == null) {
			logger.warning("Property gridName is not set! Proper clean up of handler will not be possible!");
		} else {
			this.gridName = tmpGridName;
		}

		try {
			init(uri);
		} catch (final Exception e) {
			throw new IllegalStateException("Failed to initialize UCS elastic scale handler", e);
		}

	}

	private void init(final String uri) throws UnknownHostException {

		fullPrefix = this.organization + "/ls-" + this.serverNamePrefix;
		setupAdmin();

		logger.info(this.toString());

		try {
			service = new Service(new URI(uri));
			service.login(username, password);
		} catch (final URISyntaxException e) {
			logger.severe(e.toString());
		} catch (final UCSServiceException e) {
			logger.severe(e.toString());
		}

		this.installer = new AgentlessInstaller();

		executor = Executors.newFixedThreadPool(2);
	}

	private LsServer launchServerForScaleOut() throws UCSServiceException {
		final List<LsServer> profiles = service.getAllServiceProfilesInOrg(this.organization);

		logger.info("Looking for a powered off associated profile");
		// look for switched off associated machine
		final LsServer poweredOffServer = checkForPoweredOffMachine(profiles);
		if (poweredOffServer != null) {
			return poweredOffServer;
		}

		// check if blade pool has available blade
		if (service.isComputePoolEmpty(this.bladePoolName)) {
			logger.info("Blade pool " + this.bladePoolName + " is empty - can't create new server");
			return null; // no blades in Pool
		}

		logger.info("Looking for an unassociated profile");
		// look for unassociated machine
		final LsServer unassociatedServer = checkForUnassociatedProfile(profiles);
		if (unassociatedServer != null) {
			return unassociatedServer;
		}

		// since the blade pool is not empty, create a new profile from the
		// template and use it
		final LsServer createdServer = createProfile(profiles);
		if (createdServer != null) {
			return createdServer;
		}

		return null;

	}

	@Override
	public void machineAdded(final Machine machine) {
		logger.info("UCSElasticScaleHandler: machineAdded - " + machine.getUid());
		synchronized (mutex) {
			if (pendingMachineDN == null) {
				return;
			}
		}

		String dn = null;
		try {
			dn = service.getServerDNFromMAC(getMACAddressesFromMachine(machine), this.macPoolName);
		} catch (final UCSServiceException e) {
			logger.log(Level.SEVERE, "Failed to get server DN from mac addresses", e);
			return;
		}

		synchronized (mutex) {
			if (pendingMachineDN == null) {
				return;
			}

			if (pendingMachineDN.equals(dn)) {
				if (logger.isLoggable(Level.INFO)) {
					logger.info("Machine " + machine.getUid() + " was launched by this handler with service profile "
							+ dn + " and has joined");
				}
				pendingMachineDN = null;
			}

		}

	}

	@Override
	public void machineRemoved(final Machine machine) {
		logger
				.info("UCSElasticScaleHandler: machineRemoved - " + machine.getUid() + ", Host: "
						+ machine.getHostName());
	}

	@Override
	public void processingUnitRemoved(final ProcessingUnit arg0) {
		logger.info("Processing Unit Removed: " + arg0.getName());
		if (!arg0.getName().equals(this.gridName)) {
			return;
		}

		// This data grid is being removed, so clean up
		cleanUp();

	}

	@Override
	public void scaleIn(final Machine machine) {
		logger.info("UCSElasticScaleHandler: scaleIn - " + machine.getUid());

		if (machine.getHostAddress().equals(this.locator)) {
			logger.info(machine.getUid() + " is the locator machine, not scaling in");
			return;
		}

		final String[] macAddresses = getMACAddressesFromMachine(machine);

		if (macAddresses.length == 0) {
			logger.warning("Did not find any MAC addresses for machine: " + machine.getHostAddress()
					+ ", can't scale in");
			return;
		}
		logger.info("Scaling in machine with MAC addresses: " + Arrays.toString(macAddresses));

		String serverDn = null;
		try {
			service.login(this.username, this.password);
			serverDn = service.getServerDNFromMAC(macAddresses, this.macPoolName);
		} catch (final UCSServiceException e) {
			logger.log(Level.SEVERE, "Error while querying for server by MAC", e);
			return;
		}

		logger.info("Server DN for machine: " + machine.getUid() + " is: " + serverDn);
		if (serverDn == null) {
			logger.severe("Did not find the service profile for this machine!");
		}

		try {
			logger.info("Shutting down: " + serverDn);
			service.shutdownServer(serverDn, true);
			logger.info("Server shutdown was initiated for: " + serverDn);
		} catch (final UCSServiceException e) {
			logger.log(Level.SEVERE, "Failed to shut down server with DN: " + serverDn, e);
		}

	}

	@Override
	public void scaleOut(final ElasticScaleHandlerContext context) {
		logger.info("UCSElasticScaleHandler: scaleOut");

		synchronized (mutex) {
			if (pendingMachineDN != null) {
				// Currently waiting for a machine to load!
				logger.info("A pending request for new machine " + pendingMachineDN + " is still Open. Ignoring.");
				return;
			}

			logger.info("Looking for a machine for scale out");
			try {
				service.login(username, password);
			} catch (final UCSServiceException e) {
				logger.log(Level.SEVERE, "Failed to log in to UCS manager", e);
				return;
			}

			try {

				final LsServer server = launchServerForScaleOut();

				if (server == null) {
					logger.warning("Could not find a server to scale out to!");
					return;
				}
				final List<String> macList = service.getMacsForServer(this.macPoolName, server.getDn());
				if (macList.size() == 0) {
					logger.warning("Could not find MAC address for server " + server.getDn() + " in MAC Pool "
							+ this.macPoolName + ". Server installation failed!");
					return;
				}
				final String[] macs = macList.toArray(new String[macList.size()]);

				startESMTask(macs, server.getDn());

				logger.info("New machine is launching. Profile: " + server.getDn());

				pendingMachineDN = server.getDn();

			} catch (final UCSServiceException e) {
				logger.severe("Failed to get machine from UCS: " + e.getMessage());
			}

		}

	}

	private void setupAdmin() {
		final AdminFactory factory = new AdminFactory();
		if ((this.group != null) && (this.group.length() > 0)) {
			factory.addGroup(group);
		} else if ((this.locator != null) && (this.locator.length() > 0)) {
			factory.addLocator(this.locator);
		}

		admin = factory.createAdmin();

		admin.getMachines().addLifecycleListener(this);
		admin.getProcessingUnits().getProcessingUnitRemoved().add(this);

		this.myIp = getMyIP(admin);
	}

	private void startESMTask(final String[] macs, final String serverDN) {
		executor.submit(new Runnable() {

			@Override
			public void run() {
				logger.info("Starting the agentless installation process on UCS server with mac: "
						+ Arrays.toString(macs));
				final InstallationDetails details = createInstallationDetails();

				try {
					installer.installOnMachineWithMAC(macs, myIp, details);
				} finally {
					// Just in case something went bad, make sure to clean up
					// the
					// pending
					// object on the handler, so that the next request will
					// start the
					// process over again.
					gsaTaskFinished(serverDN);
				}

			}
		});

	}

	@Override
	public String toString() {
		return "UCSElasticScaleHandler [username=" + username + ", password=" + password + ", sshUsername="
				+ sshUsername + ", sshPassword=" + sshPassword + ", gridName=" + gridName + ", bladePoolName="
				+ bladePoolName + ", fullPrefix=" + fullPrefix + ", group=" + group
				+ ", localDir=" + localDir + ", locator=" + locator + ", macPoolName=" + macPoolName + ", myIp=" + myIp
				+ ", organization=" + organization + ", remoteDir=" + remoteDir + ", serverNamePrefix="
				+ serverNamePrefix + ", template=" + template + "]";
	}

}
