package com.gigaspaces.rackspace.esm;

import java.io.File;
import java.util.Random;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.jclouds.rackspace.cloudservers.domain.Server;
import org.openspaces.admin.Admin;
import org.openspaces.admin.AdminFactory;
import org.openspaces.admin.machine.Machine;
import org.openspaces.admin.machine.events.MachineLifecycleEventListener;
import org.openspaces.admin.pu.ProcessingUnit;
import org.openspaces.admin.pu.events.ProcessingUnitRemovedEventListener;
import org.openspaces.grid.esm.ElasticScaleHandler;
import org.openspaces.grid.esm.ElasticScaleHandlerConfig;
import org.openspaces.grid.esm.ElasticScaleHandlerContext;

import com.gigaspaces.rackspace.Deployer;

/************
 * This is the implementation of the Gigaspaces Elastic Handler for the
 * Rackspace cloud. It uses the jclouds and Rackspace API to start and stop
 * servers. Note the implementations of the scaleIn and scaleOut methods.
 * 
 * @author barakme
 * 
 */
public class RackspaceElasticHandler implements ElasticScaleHandler, MachineLifecycleEventListener,
		ProcessingUnitRemovedEventListener {

	private static final int ESM_NAME_RANDOM_LIMIT = 1000;

	private static final Logger logger = 
		Logger.getLogger(RackspaceElasticHandler.class.getName());

	private Deployer deployer;

	private String lusIP;
	private final AtomicInteger serverCounter = new AtomicInteger(1);
	private String machineNamePrefix;

	private Admin admin;

	private String gridName = "SOME_RANDOM_NAME";

	@Override
	public boolean accept(final Machine arg0) {
		return true;
	}

	@Override
	public void init(final ElasticScaleHandlerConfig config) {
		
		logger.info("ESM Handler INIT. Working directory is: " + new File(".").getAbsolutePath());

		final String user = config.getProperty("user");
		final String apiKey = config.getProperty("apiKei");
		final String tmpGridName = config.getProperty("gridName");
		if(tmpGridName == null) {
			logger.warning("Property gridName is null! Proper clean up of handler will not be possible!");
		} else {
			this.gridName  = tmpGridName;
		}

		String prefix = config.getProperty("machineNamePrefix");

		if ((prefix == null) || (prefix.length() == 0)) {
			prefix = "gs-esm-gsa-";
		}

		// attach a random number to the prefix to prevent collisions
		this.machineNamePrefix = prefix + new Random().nextInt(ESM_NAME_RANDOM_LIMIT) + "-";

		logger
				.info("Creating NEW rackspace deployer with user: " + user + " and apiKey: "
						+ apiKey);

		try {
			deployer = new Deployer(user, apiKey);
		} catch (final Exception e) {
			logger.log(Level.SEVERE, "Failed to initialize Deployer", e);
			throw new IllegalStateException("Failed to initialize Rackspace Deployer: "
					+ e.getMessage(), e);
		}

		admin = new AdminFactory().createAdmin();

		// make sure I use the same IP as the ESM
		this.lusIP = admin.getElasticServiceManagers().waitForAtLeastOne().getMachine()
				.getHostAddress();

		logger.info("ESM INIT: My (LUS) IP is: " + lusIP);

		// register for machine lifecycle events so I know when machines are up
		admin.getMachines().addLifecycleListener(this);
		admin.getProcessingUnits().getProcessingUnitRemoved().add(this);

	}

	@Override
	public void machineAdded(final Machine arg0) {
		logger.info("machine added: " + arg0.getHostAddress());

	}	

	@Override
	public void machineRemoved(final Machine arg0) {
		logger.info("machine removed: " + arg0.getHostAddress());
	}

	@Override
	public void processingUnitRemoved(final ProcessingUnit arg0) {
		logger.info("Processing Unit Removed: " + arg0.getName());
		if(!arg0.getName().equals(this.gridName)) {
			return;
		}
		
		// This data grid is being removed, so clean up 
		cleanUp();

	}
	private void cleanUp() {
		
		admin.getMachines().removeLifeycleListener(this);
		admin.getProcessingUnits().getProcessingUnitRemoved().remove(this);
		deployer.destroy();
		admin.close();
		
	}

	@Override
	public void scaleIn(final Machine machine) {

		final String ip = machine.getHostAddress();

		logger.info("Scale IN -- " + ip + " --");
		
		logger.fine("Check that we are not shutting down LUS or ESM - lusIP = " + this.lusIP);
		if (this.lusIP.equals(ip)) {
			logger.info("Recieved scale in request for LUS/ESM server. Ignoring.");
			return;
		}

		logger.fine("Looking Up Racksapce server with this IP");
		final Server server = deployer.getServerWithIP(ip);
		if (server != null) {
			logger.info("Found server: " + server.getId() + ". Shutting it down");
			if (!deployer.shutdownMachine(server.getId())) {
				logger.log(Level.SEVERE, "Failed to shut down machine with Rackspace ID "
						+ server.getId());
			} else {
				logger.info("Server: " + server.getId() + " shutdown has started.");
			}
		} else {
			logger.log(Level.SEVERE, "Recieved scale in request for machine with ip " + ip + " "
					+ "and id " + machine.getUid() + " but this IP "
					+ "could not be found in the Rackspace server list");
		}

	}

	
	@Override
	public void scaleOut(final ElasticScaleHandlerContext arg0) {
		logger.info("Scale Out! Launching new server with lusIP: " + lusIP);
		final Server server = deployer.launchGSAMachine(lusIP, machineNamePrefix
				+ serverCounter.getAndIncrement(), admin);

		if (server == null) {
			logger.log(Level.SEVERE, "Failed to launch new rackspace server");
			return;
		}

		final String privateIp = (String) server.getAddresses().getPrivateAddresses().toArray()[0];
		logger.info("Launched new machine with internal IP: " + privateIp + " and ID: "
				+ server.getId());
		
	}

}
